# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The JSON schema parser validates a json data structure and if possible casts
data to the correct type and fills out default values. The result in a valid
document that can be used to construct objects.
"""

import collections
import copy
import json
import os
import re
import urllib
from urlparse import urlparse, urlunparse

import jsonschema.validators
from jsonschema._format import FormatChecker
from jsonschema.exceptions import ValidationError
from jsonschema.compat import iteritems

import fastr
from fastr import exceptions


class FastrRefResolver(jsonschema.RefResolver):
    """
    Adapted version of the RefResolver for handling inter-file references
    more to our liking
    """
    def __init__(self, base_uri, referrer, store=(), cache_remote=True, handlers=()):
        """
        Create a new FastrRefResolver

        :param str base_uri: URI of the referring document
        :param referrer: the actual referring document
        :param dict store: a mapping from URIs to documents to cache
        :param bool cache_remote: whether remote refs should be cached after
            first resolution
        :param dict handlers: a mapping from URI schemes to functions that
            should be used to retrieve them
        """

        handlers = dict(handlers)
        base_uri = urlunparse(['file', '', urllib.pathname2url(base_uri), '', '', ''])
        default_handlers = {'': FastrRefResolver.readfile,
                            'fastr': FastrRefResolver.readfastrschema}
        default_handlers.update(handlers)
        super(FastrRefResolver, self).__init__(base_uri, referrer, store, cache_remote, default_handlers)

    @classmethod
    def from_schema(cls, schema, *args, **kwargs):
        """
        Instantiate a RefResolver based on a schema
        """
        default_handlers = {'': FastrRefResolver.readfile,
                            'fastr': FastrRefResolver.readfastrschema}
        if 'handlers' in kwargs:
            handlers = dict(kwargs['handlers'])
            default_handlers.update(handlers)
        kwargs['handlers'] = default_handlers

        return cls(schema.get("id", ""), schema, *args, **kwargs)

    @staticmethod
    def readfile(filename):
        """
        Open a json file based on a simple filename

        :param str filename: the path of the file to read
        :return: the resulting json schema data
        """
        with open(filename, 'r') as fin:
            result = json.load(fin)

        return result

    @staticmethod
    def readfastrschema(name):
        """
        Open a json file based on a fastr:// url that points to a file in the
        fastr.schemadir

        :param str name: the url of the file to open
        :return: the resulting json schema data
        """
        path = urlparse(name).path[1:]
        filename = os.path.join(fastr.config.schemadir, path)
        return FastrRefResolver.readfile(filename)


def _str_to_boolean(value):
    """
    Converts a string to boolean

    :param str value: the string to convert
    :return: the result
    :rtype: bool
    """
    if isinstance(value, (str, unicode)):
        return value.lower() in ['1', 'true']
    else:
        raise TypeError('Expected a string to convert to a boolean, got a {}'.format(type(value).__name__))


def _str_to_list(value):
    """
    Converts a str to a list

    :param str value: string to convert
    :return: the resulting list
    :rtype: list
    """
    return [value]


def _default_to_object(value):
    """
    Converts anything to a dict (json object)
    :param value: the value to convert
    :return: the result
    :rtype: dict
    :raises FastrValueError: if the value cannot be converted to a dict in a sensible way
    """
    if isinstance(value, (dict, collections.Mapping)):
        return dict(value)
    elif hasattr(value, '__getstate__'):
        return value.__getstate__()
    elif hasattr(value, '__dict__'):
        return value.__dict__
    else:
        try:
            return dict(value)
        except ValueError:
            raise exceptions.FastrValueError('Cannot cast {} to a dict'.format(type(value).__name__))


def _ordereddict_to_array(value):
    """
    Specific conversion from the OrderedDict to a list (json array)
    :param OrderedDict value: the value to convert
    :return: the result
    :rtype: list
    """
    return value.values()


def _none_to_array(_):
    """
    Convert a None (json null) to a list (json array)

    :param NoneType _: the value (that is ignored)
    :return: an empty list
    :rtype: list
    """
    return []


# Default casts to get to a datatype
_DEFAULT_TYPECASTS = {
    u"array": list,
    u"boolean": bool,
    u"integer": int,
    u"number": float,
    u"object": _default_to_object,
    u"string": str,
}


# Specific casts to get to a datatype, given by ("targettype", "inputtype"): cast_func
_TYPECASTS = {
    (u"array", collections.OrderedDict): _ordereddict_to_array,
    (u"array", type(None)): _none_to_array,
    (u"array", str): _str_to_list,
    (u"boolean", str): _str_to_boolean,
    (u"boolean", unicode): _str_to_boolean,
}


def _refer(validator, schema):
    """
    Follow a references in the schema

    :param validator: the json schema validator
    :param dict schema: the current json schema
    :return: the new json schema
    :rtype: dict
    """
    ref = schema.get(u"$ref")
    if ref is None:
        return schema

    with validator.resolver.resolving(ref) as resolved:
        result = resolved

        if '$ref' in result:
            return _refer(validator, result)
        else:
            return result


def _match_type(validator, instance, schema):
    """
    Match the datatype of and instance with a datatype required by the schema.

    :param validator: Validator used
    :param instance: The instance of the data
    :param dict schema: The schema describing the instance
    :return: the instance in the matched type
    """
    desired_type = schema.get('type')
    if not isinstance(desired_type, list):
        desired_type = [desired_type]

    if not any(validator.is_type(instance, dtype) for dtype in desired_type):
        for dtype in desired_type:
            try:
                if (dtype, type(instance)) in _TYPECASTS:
                    return _TYPECASTS[dtype, type(instance)](instance)
                return _DEFAULT_TYPECASTS[dtype](instance)
            except (ValueError, TypeError):
                pass  # Not coercable, just hope another will success
        #raise TypeError('Cannot match the data and schema type ({} and {})!'.format(type(instance).__name__, desired_type))
    else:
        if isinstance(instance, unicode):
            return str(instance)
        else:
            return instance


def pattern_properties_prevalid(validator, pattern_properties, instance, schema):
    """
    The pre-validation function for patternProperties

    :param validator: the json schema validator
    :param dict pattern_properties: the current patternProperties
    :param dict instance: the current object instance
    :param dict schema: the current json schema
    """
    if not validator.is_type(instance, "object"):
        return

    properties = set(schema.get('properties', {}).keys())
    pattern_properties = _refer(validator, pattern_properties)

    for key, _ in instance.items():
        if key in properties:
            continue

        for pattern, subschema in iteritems(pattern_properties):
            if not re.match(pattern, key):
                continue

            subschema = _refer(validator, subschema)
            instance[key] = _match_type(validator, instance[key], subschema)

            break


def properties_prevalidate(validator, properties, instance, schema):
    """
    The pre-validation function for properties

    :param validator: the json schema validator
    :param dict properties: the current properties
    :param instance: the current object instance
    :param dict schema: the current json schema
    """
    # All arguments must be used because this function is called like this
    # pylint: disable=unused-argument
    if not validator.is_type(instance, "object"):
        return

    properties = _refer(validator, properties)

    for property_name, subschema in iteritems(properties):
        subschema = _refer(validator, subschema)
        if property_name in instance:
            instance[property_name] = _match_type(validator, instance[property_name], subschema)


def items_prevalidate(validator, items, instance, schema):
    """
    The pre-validation function for items

    :param validator: the json schema validator
    :param dict items: the current items
    :param instance: the current object instance
    :param dict schema: the current json schema
    """
    # All arguments must be used because this function is called like this
    # pylint: disable=unused-argument
    if instance is None:
        return

    if isinstance(items, dict):
        subschema = _refer(validator, items)
        for idx, item in enumerate(instance):
            instance[idx] = _match_type(validator, item, subschema)
    elif isinstance(items, list) and len(items) == len(instance):
        for idx, (item, subschema) in enumerate(zip(instance, items)):
            subschema = _refer(validator, subschema)
            instance[idx] = _match_type(validator, item, subschema)
    else:
        raise ValueError('Expected a list or a tuple notation, found neither')


def properties_postvalidate(validator, properties, instance, schema):
    """
    # All arguments must be used because this function is called like this
    # pylint: disable=unused-argument
    The post-validation function for properties

    :param validator: the json schema validator
    :param dict properties: the current properties
    :param instance: the current object instance
    :param dict schema: the current json schema
    """
    # All arguments must be used because this function is called like this
    # pylint: disable=unused-argument
    for property_name, subschema in iteritems(properties):
        if instance is None:
            print('wtf? {} / {} -> {}'.format(instance, property_name, subschema))
        if property_name not in instance and "default" in subschema:
            instance[property_name] = subschema["default"]


def one_of_draft4(validator, one_of, instance, schema):
    """
    The one_of directory needs to be done stepwise, because a validation
    even if it fails will try to change types / set defaults etc. Therefore
    we first create a copy of the data per subschema and test if they match.
    Once we found a proper match, we only validate that branch on the real data
    so that only the valid piece of schema will effect the data.

    :param validator: the json schema validator
    :param dict one_of: the current one_of
    :param instance: the current object instance
    :param dict schema: the current json schema
    """
    # All arguments must be used because this function is called like this
    # pylint: disable=unused-argument
    subschemas = enumerate(one_of)
    all_errors = []
    first_valid = {}

    for index, subschema in subschemas:
        temp_instance = copy.deepcopy(instance)
        errs = list(validator.descend(temp_instance, subschema, schema_path=index))
        if not errs:
            first_valid = subschema
            break
        all_errors.extend(errs)
    else:
        # Make sure the reference is available later
        yield ValidationError(
            "%r is not valid under any of the given schemas" % (instance,),
            context=all_errors,
            )

    more_valid = []
    for _, subschema in subschemas:
        temp_instance = copy.deepcopy(instance)
        if validator.is_valid(temp_instance, subschema):
            more_valid.append(subschema)

    if more_valid:
        more_valid.append(first_valid)
        reprs = ", ".join(repr(schema) for schema in more_valid)
        yield ValidationError(
            "%r is valid under each of %s" % (instance, reprs)
        )

    validator.validate(instance, first_valid)


def any_of_draft4(validator, any_of, instance, schema):
    """
    The oneOf directory needs to be done stepwise, because a validation
    even if it fails will try to change types / set defaults etc. Therefore
    we first create a copy of the data per subschema and test if they match.
    Then for all the schemas that are valid, we perform the validation on the
    actual data so that only the valid subschemas will effect the data.

    :param validator: the json schema validator
    :param dict any_of: the current oneOf
    :param instance: the current object instance
    :param dict schema: the current json schema
    """
    # All arguments must be used because this function is called like this
    # pylint: disable=unused-argument
    all_errors = []
    valid_subschemas = []
    for index, subschema in enumerate(any_of):
        temp_instance = copy.deepcopy(instance)
        errs = list(validator.descend(temp_instance, subschema, schema_path=index))

        if not errs:
            valid_subschemas.append((index, subschema))
        else:
            all_errors.extend(errs)

    if len(valid_subschemas) == 0:
        yield ValidationError(
            "%r is not valid under any of the given schemas" % (instance,),
            context=all_errors,
            )
    else:
        for index, subschema in valid_subschemas:
            validator.validate(instance, subschema)


def not_draft4(validator, not_schema, instance, schema):
    """
    The not needs to use a temporary copy of the instance, not to change the
    instance with the invalid schema

    :param validator: the json schema validator
    :param dict not_schema: the current oneOf
    :param instance: the current object instance
    :param dict schema: the current json schema
    """
    # All arguments must be used because this function is called like this
    # pylint: disable=unused-argument
    temp_instance = copy.deepcopy(instance)  # Make sure not to change instance

    if validator.is_valid(temp_instance, not_schema):
        yield ValidationError(
            "%r is not allowed for %r" % (not_schema, instance)
        )


def extend(validator_cls):
    """
Extend the given :class:`jsonschema.IValidator` with the Seep layer.

"""

    Validator = jsonschema.validators.extend(
        validator_cls, {
            u"anyOf": any_of_draft4,
            u"oneOf": one_of_draft4,
            u"not": not_draft4
            }
    )

    class Blueprinter(Validator):
        PREVALIDATORS = collections.OrderedDict()
        POSTVALIDATORS = collections.OrderedDict()

        def __init__(self, uri, schema, types=(), resolver=None, format_checker=None):
            if resolver is None:
                resolver = FastrRefResolver(uri, schema)

            if format_checker is None:
                format_checker = FormatChecker()

            super(Blueprinter, self).__init__(schema=schema,
                                              types=types,
                                              resolver=resolver,
                                              format_checker=format_checker)
            self._stack = []
            self.network = None

        def instantiate(self, data, network=None):
            result = [data]
            self.network = network
            self._stack.append(result)
            self.validate(data)

            if len(result) != 1:
                raise ValueError('Something went wrong!')

            self.network = None
            return result[0]

        def iter_errors(self, instance, _schema=None):
            if _schema is None:
                _schema = self.schema

            with self.resolver.in_scope(_schema.get("id", "")):
                self._stack.append(instance)

                ref = _schema.get("$ref")
                if ref is not None:
                    validators = [("$ref", ref)]
                else:
                    validators = iteritems(_schema)

                # Iterate over PREVALIDATORS so we can control their order
                for k, action in self.PREVALIDATORS.items():
                    if k in _schema:
                        action(self, _schema[k], instance, _schema)

                errors = []
                for k, value in validators:
                    validator = self.VALIDATORS.get(k)
                    if validator is None:
                        continue

                    extra_errors = tuple(validator(self, value, instance, _schema)) or ()

                    for error in extra_errors:
                        # set details if not already set by the called fn
                        error._set(
                            validator=k,
                            validator_value=value,
                            instance=instance,
                            schema=_schema,
                            )
                        if k != "$ref":
                            error.schema_path.appendleft(k)
                        errors.append(error)

                self._stack.pop()

                for error in errors:
                    yield error

                # Iterate over POSTVALIDATORS so we can control their order
                for k, action in self.POSTVALIDATORS.items():
                    if k in _schema:
                        action(self, _schema[k], instance, _schema)

    Blueprinter.PREVALIDATORS['properties'] = properties_prevalidate
    Blueprinter.PREVALIDATORS['patternProperties'] = pattern_properties_prevalid
    Blueprinter.PREVALIDATORS['items'] = items_prevalidate
    Blueprinter.POSTVALIDATORS['properties'] = properties_postvalidate
    return Blueprinter


def getblueprinter(uri, blueprint=None):
    """
Instantiate the given data using the blueprinter.

:argument blueprint: a blueprint (JSON Schema with Seep properties)

"""

    if blueprint is None:
        with open(uri, 'r') as fin:
            try:
                blueprint = json.load(fin)
            except ValueError as exception:
                raise ValueError('{} ({})'.format(exception.message, uri))

    Validator = jsonschema.validators.validator_for(blueprint)
    blueprinter = extend(Validator)(uri, blueprint)
    return blueprinter



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Manager class for Plugins in the fastr system
"""

import collections
import imp
import inspect
import os
import sys
import traceback
from abc import abstractproperty
from types import ModuleType

import fastr
from fastr.core.basemanager import BaseManager
from fastr.core.baseplugin import BasePlugin, Plugin, PluginState
import fastr.exceptions as exceptions


plugin_option_type = collections.namedtuple('plugin_option_type', ['filename', 'name', 'namespace', 'id'])


class BasePluginManager(BaseManager):
    """
    Baseclass for PluginManagers, need to override the self._plugin_class
    """

    def __init__(self, path=None, recursive=False):
        """
        Create a BasePluginManager and scan the give path for matching plugins

        :param str path: path to scan
        :param bool recursive: flag to indicate a recursive search
        :return: newly created plugin manager
        :raises FastrTypeError: if self._plugin_class is set to a class not
                                 subclassing BasePlugin
        """
        self._loaded_plugins = {}
        self._plugin_options = {}

        super(BasePluginManager, self).__init__(path, recursive)

        if not issubclass(self.plugin_class, BasePlugin):
            raise exceptions.FastrTypeError('Plugin type to manage ({}) not a valid plugin! (needs to be subclass of BasePlugin)'.format(self.plugin_class.__name__))

    def __getitem__(self, key):
        """
        Retrieve item from BaseManager

        :param key: the key of the item to retrieve
        :return: the value indicated by the key
        :raises FastrKeyError: if the key is not found in the BaseManager
        """
        try:
            return super(BasePluginManager, self).__getitem__(key)
        except exceptions.FastrKeyError:
            self.load_plugin(key.lower())

            return super(BasePluginManager, self).__getitem__(key)

    @abstractproperty
    def plugin_class(self):
        """
        The class from which the plugins must be subclassed
        """
        raise exceptions.FastrNotImplementedError

    @property
    def _item_extension(self):
        """
        Plugins should be loaded from files with a .py extension
        """
        return '.py'

    @property
    def _instantiate(self):
        """
        Flag indicating that the plugin should be instantiated prior to saving
        """
        return True

    def _print_key(self, key):
        print_out = (self[key].status.value, key)

        return print_out

    def _print_value(self, val):
        """
        Function for printing values (plugins) in this manager

        :param BasePlugin val: value to print
        :return: print representation
        :rtype: str
        """
        if val._instantiate:
            val = type(val)

        print_out = '<{}: {}>'.format(val.__bases__[0].__name__, val.__name__)
        return print_out

    def _load_item(self, filepath, namespace):
        """
        Load a plugin

        :param str filepath: path of the plugin to load
        """
        name = os.path.basename(filepath)
        name = os.path.splitext(name)[0].lower()

        value = plugin_option_type(filename=filepath, name=name, namespace=namespace, id=None)

        self._plugin_options[name] = value

    def populate(self):
        """
        Populate the manager with the data. This is a method that will be
        called when the Managers data is first accessed. This way we avoid
        doing expensive directory scans when the data is never requested.
        """
        super(BasePluginManager, self).populate()

        for plugin_key in self._plugin_options.keys():
            if plugin_key in self._plugin_options:
                self.load_plugin(plugin_key)

        if 'VirtualFileSystem' not in self:
            self._store_item('VirtualFileSystem', fastr.vfs)

    def load_plugin(self, plugin_key):
        plugin_option = self._plugin_options[plugin_key]
        filepath = plugin_option.filename

        # Since we cannot know what Plugins might throw, catch all
        # pylint: disable=broad-except
        try:
            filebase, _ = os.path.splitext(os.path.basename(filepath))
            temp_module = imp.load_source(filebase, filepath)
            for name, obj in inspect.getmembers(temp_module):
                if inspect.isclass(obj):
                    if filebase.lower() != obj.__name__.lower():
                        fastr.log.debug('Plugin name and module do not match ({} vs {})'.format(obj.__name__, filebase))
                        continue

                    if not issubclass(obj, self.plugin_class):
                        fastr.log.debug('{} is not a subclass of {}'.format(obj, self.plugin_class))
                        continue

                    obj.filename = filepath
                    if not inspect.isabstract(obj):
                        if obj.status == PluginState.uninitialized:
                            # Since we cannot know what Plugins might throw, catch all
                            # pylint: disable=broad-except
                            try:
                                # Register the configuration for the plugin
                                obj.register_configuration()

                                obj.set_status(PluginState.preload, 'Set to PreLoad to perform testing')  # Let the Plugin think it is loaded, or it will refuse to instantiate
                                obj.test()
                                obj.set_status(PluginState.loaded, 'Testing successful, loaded properly')  # Let the Plugin think it is loaded, or it will refuse to instantiate
                            except Exception as exception:
                                fastr.log.warning('Could not load plugin file {}\n{}'.format(filepath, exception))
                                exc_type, _, _ = sys.exc_info()
                                exc_info = traceback.format_exc()
                                fastr.log.debug('Encountered exception ({}) during instantiation of the plugin:\n{}'.format(exc_type.__name__, exc_info))
                                exception_stacktrace = ('Encountered exception ({}) during'
                                                        ' instantiation of the plugin:\n{}').format(exc_type.__name__, exc_info)
                                exception_message = '[{}] {}'.format(obj.fullid, exception.message)
                                obj.set_status(PluginState.failed, exception_message, exception_stacktrace)
                        elif obj.status not in (PluginState.loaded, PluginState.failed):
                            fastr.log.warning('Invalid Plugin status: {}!'.format(obj.status))

                        # Save the source in the obj
                        obj.set_code(inspect.getsource(obj))
                        obj.module = temp_module

                        if obj.instantiate:
                            fastr.log.debug('Store instantiated plugin')
                            self._store_item(name, obj())
                        else:
                            fastr.log.debug('Store uninstantiated plugin')
                            self._store_item(name, obj)

                    else:
                        fastr.log.debug('Skipping abstract Plugin: {} ({})'.format(name, filepath))

        except Exception as exception:
            fastr.log.warning('Could not load {} file {}\n{}'.format(self.plugin_class.__name__, filepath, exception))
            exc_type, _, _ = sys.exc_info()
            exc_info = traceback.format_exc()
            fastr.log.info('Encountered exception ({}) during loading of the plugin:\n{}'.format(exc_type.__name__, exc_info))
        finally:
            del self._plugin_options[plugin_key]


class PluginsView(collections.MutableMapping):
    """
    A collection that acts like view of the plugins of another plugin manager.
    This is a proxy object that only gives access the plugins of a certain
    plugin class. It behaves like a mapping and is used as the data object for
    a PluginSubManager.
    """
    def __init__(self, parent, plugin_class):
        """
        Constructor for the plugins view

        :param BasePluginManager parent: the parent plugin manager
        :param class plugin_class: the class of the plugins to expose
        """
        self.plugin_class = plugin_class
        self.parent = parent

    def filter_plugin(self, plugin):
        if self.plugin_class.instantiate and isinstance(plugin, self.plugin_class):
            return True
        elif not self.plugin_class.instantiate and issubclass(plugin, self.plugin_class):
            return True
        else:
            return False

    def __getitem__(self, item):
        result = self.parent[item]

        if not self.filter_plugin(plugin=result):
            raise KeyError(item)

        return result

    def __setitem__(self, key, value):
        if not self.filter_plugin(value):
            raise TypeError(value)

        if key in self.parent:
            if not self.filter_plugin(self.parent[key]):
                raise TypeError(key)

        self.parent[key] = value

    def __delitem__(self, key):
        if key in self.parent:
            if not self.filter_plugin(self.parent[key]):
                raise TypeError(key)

        del self.parent[key]

    def __len__(self):
        return sum(1 for v in self.parent.values() if self.filter_plugin(v))

    def __iter__(self):
        for key, value in self.parent.items():
            if not self.filter_plugin(value):
                continue

            yield key


class PluginSubManager(BasePluginManager):
    """
    A PluginManager that is a selection of a parent plugin manger. It uses the
    PluginsView to only exponse part of the parent PluginManager. This is used
    to create plugin managers for only certain types of plugins (e.g. IOPlugins)
    without loading them multiple times.
    """
    def __init__(self, parent, plugin_class):
        self.parent = parent
        self._plugin_class = plugin_class
        self._data_link = PluginsView(parent=parent, plugin_class=plugin_class)
        super(PluginSubManager, self).__init__()

    @property
    def data(self):
        return self._data_link

    @property
    def plugin_class(self):
        """
        PluginSubManagers only expose the plugins of a certain class
        """
        return self._plugin_class


class PluginManager(BasePluginManager):
    def __init__(self, path=None):
        if path is None:
            path = fastr.config.plugins_path

        super(PluginManager, self).__init__(path=path, recursive=True)

    @property
    def plugin_class(self):
        """
        The plugin manager contains any Plugin subclass
        """
        return Plugin

    def __setitem__(self, key, value):
        """
        Store an item in the BaseManager, will ignore the item if the key is
        already present in the BaseManager.

        :param name: the key of the item to save
        :param value: the value of the item to save
        :return: None
        """
        if not (isinstance(value, Plugin) or issubclass(value, Plugin)):
            raise TypeError(value)

        super(PluginManager, self).__setitem__(key, value)

    def _store_item(self, name, value):
        """
        Store an item in the BaseManager, will ignore the item if the key is
        already present in the BaseManager.

        :param name: the key of the item to save
        :param value: the value of the item to save
        :return: None
        """
        if name in self.keys():
            fastr.log.warning('Skipping {} from {} (the plugin is already in the {})'.format(name, value.filename, type(self).__name__))
        else:

            # Set the module to the fastr plugins
            if isinstance(value, BasePlugin):
                type(value).__module__ = 'fastr.plugins'
                setattr(fastr.plugins, value.id, type(value))
            else:
                value.__module__ = 'fastr.plugins'
                setattr(fastr.plugins, value.id, value)

            self[name] = value


class LazyModule(ModuleType):
    """
    A module that allows content to be loaded lazily from plugins. It generally
    is (almost) empty and gets (partially) populated when an attribute cannot be
    found. This allows lazy loading and plugins depending on other plugins.
    """
    def __init__(self, name, parent, plugin_manager):
        super(LazyModule, self).__init__(name)
        self._plugin_manager = plugin_manager
        self.__dict__.update(vars(parent))
        self._parent = parent
        sys.modules[parent.__name__] = self

    def __repr__(self):
        return super(LazyModule, self).__repr__().replace('module', 'lazy_module', 1)

    def __getattr__(self, item):
        """
        The getattr is called when getattribute does not return a value and is
        used as a fallback. In this case we try to find the value normally and
        will trigger the plugin manager if it cannot be found.

        :param str item: attribute to retrieve
        :return: the requested attribute
        """
        try:
            return super(LazyModule, self).__getattribute__(item)
        except AttributeError as exception:
            try:
                return self._plugin_manager[item]
            except KeyError:
                raise exception



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains tool for converting python dictionaries into XML object
and vice-versa.
"""
__author__ = 'hachterberg'

import re
import xml.etree.ElementTree as ElementTree
import xml.dom.minidom
import fastr

__list_entry_tag__ = 'li'
__tag_symbol_start__ = 'p_'


def _toxml(data, element=None, key=None):
    """
    Write data to an XML string

    :param data: object to write
    :param element: element to write
    :param key: key to set
    :return: XML string
    :rtype: str
    """
    toplevel = False
    if element is None:
        toplevel = True
        if key is None:
            key = 'root'

    if isinstance(data, dict):
        if element is not None:
            subelement = ElementTree.SubElement(element, key)
        else:
            element = ElementTree.Element(key)
            subelement = element

        for key, value in data.items():
            key = _prepkey(key)

            if _issimpletype(value):
                if isinstance(value, (str, unicode)) and ("\n" in value or len(value) > 48):
                    extra_element = ElementTree.SubElement(subelement, key)
                    extra_element.text = _to_str(value)
                else:
                    subelement.set(key, _to_str(value))
            elif value is None:
                ElementTree.SubElement(subelement, key)
            elif isinstance(value, (list, tuple)):
                _toxml(value, subelement, key)
            elif isinstance(value, dict):
                _toxml(value, subelement, key)
            else:
                raise TypeError('Invalid class to serialize ({})'.format(type(value).__name__))

    elif isinstance(data, (list, tuple)):
        if element is not None:
            subelement = ElementTree.SubElement(element, key)
        else:
            element = ElementTree.Element(key)
            subelement = element

        if key[-1] == 's':
            subkey = key[:-1]
        else:
            subkey = __list_entry_tag__

        for value in data:
            if _issimpletype(value):
                entry_element = ElementTree.SubElement(subelement, subkey)
                entry_element.text = _to_str(value)
            elif isinstance(value, list):
                _toxml(value, subelement, subkey)
            elif isinstance(value, dict):
                _toxml(value, subelement, subkey)
            elif value is None:
                ElementTree.SubElement(subelement, subkey)
            else:
                raise TypeError('Invalid class to serialize ({})'.format(type(value).__name__))

    elif data is None:
        if key is not None:
            element.set(key, '')

    elif _issimpletype(data):
        if key is not None:
            element.set(key, _to_str(data))
        else:
            element.text = _to_str(data)

    else:
        fastr.log.warning('Warning invalid type {}'.format(type(data).__name__))

    if toplevel:
        xml_string = xml.dom.minidom.parseString(ElementTree.tostring(element))
        xml_string = xml_string.toprettyxml(indent='    ').encode('utf-8')
        return xml_string


def _fromxml(root):
    """
    Parse ETree into objects

    :param root: root elemnt to parse
    :return:
    """
    if root.tag == 'root':
        return _parse_elem(root)
    else:
        return _parse_elem(root)


def _issimpletype(data):
    """
    Check if a value is a simple type (str, int, float, bool, unicode)

    :param data: value to check
    :return: flag indicating a simple type
    """
    return isinstance(data, (str, int, float, bool, unicode))


def _prepkey(key):
    """
    Prepare a key for use in XML, this means padding it with valid characters
    if needed and encoding special characters.

    :param str key: key to prepare
    :return: prepared key
    :rtype: str
    """
    newstr = []
    for char in key:
        if char.isalnum() or char in '-_':
            newstr.append(char)
        else:
            newstr.append('__{:03d}__'.format(ord(char)))

    key = ''.join(newstr)

    if not key[0].isalpha():
        key = '{}{}'.format(__tag_symbol_start__, key)
    return key


def _parse_elem(element):
    """
    Parse a single element

    :param element: element to parse
    :return: resulting object
    """
    # Prepare text, avoid whitespace issues
    text = element.text
    if text is not None and text.strip() == '':
        text = None

    if text is None and len(element.attrib) == 0 and len(element) == 0:
        # Empty, must be a none
        return None
    elif text is None and len(element.attrib) == 0 and all([x.tag == __list_entry_tag__ or x.tag == element.tag[:-1] for x in element]):
        # Must be list
        return [_parse_elem(x) for x in element]
    elif text is not None:
        # Must be simple text
        return _from_str(element.text)
    else:
        # Must be dict
        result = dict()

        for key, value in element.attrib.items():
            key = _parse_key(key)
            result[key] = _from_str(value)

        # Check if we have list children
        counts = {}
        for child in element:
            key = _parse_key(child.tag)
            if key in counts:
                counts[key] += 1
            else:
                counts[key] = 1

        for child in element:
            key = _parse_key(child.tag)

            if counts[key] == 1:
                # Entry is a single value child
                result[key] = _parse_elem(child)
            else:
                # An entry is a list
                if key not in result:
                    result[key] = []
                result[key].append(_parse_elem(child))

        return result


def _parse_key(key):
    """
    Parse a key, removes padding an special character encoding used for XML

    :param str key: key to parse
    :return: cleaned key
    :rtype: str
    """
    if key.startswith(__tag_symbol_start__):
        key = key[len(__tag_symbol_start__):]

    for match in re.findall(r'__\d\d\d__', key):
        key = re.sub(match, chr(int(match[2:5])), key)

    return key


def _to_str(value):
    """
    Convert a basic type to a str representation

    :param value: value to convert
    :return: string version
    :rtype: str
    """
    if isinstance(value, str):
        try:
            float(value)
            return '___{}'.format(value)
        except ValueError:
            return value
    else:
        return str(value)


def _from_str(text):
    """
    Convert a str to another type

    :param str text: str to parse
    :return: parsed value
    """
    # Convert to Boolean
    if text.lower() == 'true':
        return True
    elif text.lower() == 'false':
        return False
    # Find strings that started with illegal character and get escaped by ___
    elif text.startswith('___'):
        return text[3:]

    # Attempt conversion to numericals
    try:
        out = int(text)
        return out
    except ValueError:
        pass

    try:
        out = float(text)
        return out
    except ValueError:
        pass

    # Must be a str
    return str(text)


def dumps(data):
    """
    Write a dict to an XML string

    :param data: data to write
    :return: the XML data
    :rtype: str
    """
    return _toxml(data)


def dump(data, filehandle):
    """
    Write a dict to an XML file

    :param data: data to write
    :param filehandle: file handle to write to
    """
    filehandle.write(dumps(data))


def loads(data):
    """
    Load an xml string and parse it to a dict

    :param str data: the xml data to load
    :return: the parsed data
    """
    root = ElementTree.fromstring(data)
    return _fromxml(root)


def load(filehandle):
    """
    Load an xml file and parse it to a dict

    :param filehandle: file handle to load
    :return: the parsed data
    """
    tree = ElementTree.parse(filehandle)
    root = tree.getroot()
    return _fromxml(root)



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import fastr
import fastr.execution.executionscript


def get_parser():
    parser = argparse.ArgumentParser(description="Execute a job or network from commandline")
    parser.add_argument('job', metavar='JOBFILE', nargs='?', type=str,
                        help='File of the job to execute (default ./__fastr_command__.pickle.gz)')
    return parser


def main(args=None, unknown_args=None):
    """
    Execute a fastr job file
    """
    if args is None and unknown_args is None:
        # No arguments were parsed yet, parse them now
        parser = get_parser()
        args, unknown_args = parser.parse_known_args()

    if args.job is not None:
        job = args.job
    else:
        curdir_contents = os.listdir('.')

        if '__fastr_command__.pickle.gz' in curdir_contents:
            job = '__fastr_command__.pickle.gz'
        else:
            fastr.log.critical('No job given and cannot find __fastr_command__.pickle.gz in current directory!')
            exit()

    fastr.execution.executionscript.main(job)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
FASTR_LOG_TYPE = 'none'  # Do not get info about fastr


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('infile', metavar='result.pickle.gz', help='result file to cat')
    parser.add_argument('path', help='path of the data to print')
    return parser


def main(args=None, unknown_args=None):
    """
    Print information from a job file
    """
    if args is None and unknown_args is None:
        # No arguments were parsed yet, parse them now
        parser = get_parser()
        args, unknown_args = parser.parse_known_args()

    import gzip
    import os
    import pickle
    from pprint import pprint

    infile = args.infile
    path = args.path.strip('/').split('/')

    if os.path.isdir(infile):
        infile = os.path.join(infile, '__fastr_result__.pickle.gz')

        if not os.path.isfile(infile):
            print('ERROR: Input file does not exist!')
            return

    try:
        with gzip.open(infile) as fh:
            job = pickle.load(fh)
    except IOError:
        print('ERROR: Could not open {} for reading'.format(infile))

    data = job.info_store

    for part in path:
        try:
            data = data[part]
        except KeyError:
            raise KeyError('{} not found in {}'.format(part, data.keys()))

    if isinstance(data, str):
        for line in data.split('\n'):
            print line
    else:
        pprint(data)


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module keeps track of the version of the currently used Fastr framework.
It can check its version from mercurial or a saved file
"""
import os

_VERSIONFILE = os.path.join(os.path.dirname(__file__), 'versioninfo')
_FASTR_SRC_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), '..'))


def get_base_version():
    """
    Get the version from the top-level version file

    :return: the version
    :rtype str:
    """
    try:
        with open(os.path.join(os.path.dirname(__file__), '..', 'version'), 'r') as version_file:
            return version_file.read().strip()
    except IOError:
        return None


def get_hg_info():
    """
    Read information about the current mercurial branch and revision

    :return: tuple containing head revision and branch
    """
    # Get information about the version (polling mercurial if possible)
    dirstate = os.path.join(_FASTR_SRC_DIR, '.hg', 'dirstate')
    if os.path.isfile(dirstate):
        with open(dirstate, 'rb') as dirstate_file:
            current_hg_head = dirstate_file.read(20).encode('hex')
    else:
        return None, None

    branch = os.path.join(_FASTR_SRC_DIR, '.hg', 'branch')
    if os.path.isfile(branch):
        with open(branch, 'r') as branch_file:
            current_hg_branch = branch_file.read().strip()
    else:
        return None, None

    return current_hg_head, current_hg_branch


def get_saved_version():
    """
    Get cached version from file

    :return: tuple with version, head revision and branch
    """
    try:
        with open(_VERSIONFILE, 'r') as version_file:
            data = version_file.read().split()
            data = dict(x.split('=', 1) for x in data)

        return data['version'], data['hg_head'], data['hg_branch']
    except IOError:
        return None, None, None


def save_version(current_version, current_hg_head, current_hg_branch):
    """
    Cache the version information (useful for when installing)

    :param str current_version: version
    :param str current_hg_head: mercurial head revision
    :param str current_hg_branch: mercurial branch
    :return:
    """
    with open(_VERSIONFILE, 'w') as version_file:
        version_file.write('version={}\n'.format(current_version))
        version_file.write('hg_head={}\n'.format(current_hg_head))
        version_file.write('hg_branch={}\n'.format(current_hg_branch))


def clear_version():
    """
    Remove the cached version info
    """
    os.remove(_VERSIONFILE)

# Try to get version info
version = get_base_version()
hg_head, hg_branch = get_hg_info()

# Note that we imported from a mercurial repository
from_mercurial = hg_branch is not None


# Try to get missing info from saved versioninfo
if version is None or hg_head is None or hg_branch is None:
    _n_version, _n_hg_head, _n_hg_branch = get_saved_version()

    if version is None:
        version = _n_version

    if hg_head is None:
        hg_head = _n_hg_head

    if hg_branch is None:
        hg_branch = _n_hg_branch

full_version = '{}_{}_{}'.format(version, hg_branch, hg_head[:7])
not_default_branch = hg_branch != 'default'



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module containing the code to create class properties.
"""

class ClassPropertyDescriptor(object):
    """
    A descriptor that can act like a property for a class.
    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, obj, cls=None):
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()


def classproperty(func):
    """
    Decorator to create a "class property"

    :param func: the function to wrap
    :return: a class property
    :rtype: ClassPropertyDescriptor
    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)





# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module holds the ExecutionPluginManager as well as the base-class
for all ExecutionPlugins.
"""

from abc import abstractmethod
import functools
import gzip
import pickle
import sys
import threading
import traceback
import types

from enum import Enum

import fastr
from fastr import exceptions
from fastr.core.baseplugin import Plugin
from fastr.core.pluginmanager import PluginSubManager
from fastr.execution.job import Job, JobState
from fastr.utils import iohelpers


class JobAction(Enum):
    """
    Job actions that can be performed. This is used for checking
    if held jobs should be queued, held longer or be cancelled.
    """

    hold = 'hold'
    queue = 'queue'
    cancel = 'cancel'


class ExecutionPlugin(Plugin):
    """
    This class is the base for all Plugins to execute jobs somewhere. There are
    many methods already in place for taking care of stuff. Most plugins should
    only need to redefine a few abstract methods:

    * :py:meth:`__init__ <fastr.execution.executionpluginmanager.ExecutionPlugin.__init__>`
      the constructor
    * :py:meth:`cleanup <fastr.execution.executionpluginmanager.ExecutionPlugin.__init__>`
      a clean up function that frees resources, closes connections, etc
    * :py:meth:`_queue_job <fastr.execution.executionpluginmanager.ExecutionPlugin._queue_job>`
      the method that queues the job for execution
    * :py:meth:`_cancel_job <fastr.execution.executionpluginmanager.ExecutionPlugin._cancel_job>`
      cancels a previously queued job
    * :py:meth:`_release_job <fastr.execution.executionpluginmanager.ExecutionPlugin._release_job>`
      releases a job that is currently held
    * :py:meth:`_job_finished <fastr.execution.executionpluginmanager.ExecutionPlugin._job_finished>`
      extra callback for when a job finishes

    Not all of the functions need to actually do anything for a plugin. There
    are examples of plugins that do not really need a ``cleanup``, but for
    safety you need to implement it. Just using a ``pass`` for the method could
    be fine in such a case.

    .. warning::

        When overwriting other function, extreme care must be taken not to break
        the plugins working.
    """

    @abstractmethod
    def __init__(self, finished_callback=None, cancelled_callback=None, status_callback=None):
        """
        Setup the ExecutionPlugin

        :param finished_callback: the callback to call after a job finished
        :param cancelled_callback: the callback to call after a job cancelled
        :return: newly created ExecutionPlugin
        """
        super(ExecutionPlugin, self).__init__()

        # Pylint seems to be unable to figure out the .dict() member
        # pylint: disable=no-member

        self.job_status = {}
        self.job_dict = {}
        self.job_archive = {}
        self._finished_callback = finished_callback
        self._cancelled_callback = cancelled_callback
        self._status_callback = status_callback

        # Dict indicating the depending jobs for a certain jobs (who is waiting on the key jobid)
        self.held_queue = {}
        self.held_queue_lock = threading.Lock()

        # Flag indicating the plugin is accepting new jobs queued
        self.accepting = True

    def __enter__(self):
        return self

    def __exit__(self, type_, value, tb):
        self.cleanup()

    def __del__(self):
        """
        Cleanup if the variable was deleted on purpose
        """
        fastr.log.debug('Calling cleanup')
        self.cleanup()

    @abstractmethod
    def cleanup(self):
        """
        Method to call to clean up the ExecutionPlugin. This can be to clear
        temporary data, close connections, etc.

        :param force: force cleanup (e.g. kill instead of join a process)
        """
        # Stop accepting new jobs (close the queue)
        self.accepting = False

        # Cancel all queued jobs
        while len(self.job_dict) > 0:
            jobid, job = self.job_dict.popitem()
            fastr.log.debug('Cleanup cancelling {}'.format(jobid))
            self.cancel_job(job)

    def queue_job(self, job):
        """
        Add a job to the execution queue

        :param Job job: job to add
        """
        if not self.accepting:
            return

        if isinstance(job, list):
            for j in job:
                self.queue_job(j)
            return

        self.job_dict[job.jobid] = job
        self.job_status[job.jobid] = job.status = JobState.queued

        for hold_id in job.hold_jobs:
            # Add job reference to held queue to receive signal when the
            # required jobs are finished/failed. Do not subscribe for jobs that
            # are already finished.
            if hold_id in self.job_status and self.job_status[hold_id].done:
                continue

            with self.held_queue_lock:
                if hold_id not in self.held_queue:
                    self.held_queue[hold_id] = []

                # append to held_queue, because of the managed dict, we need to replace the value (not update)
                self.held_queue[hold_id].append(job.jobid)

        # Save the job to file before serializing
        with gzip.open(job.commandfile, 'wb') as fout:
            fastr.log.debug('Writing job pickle.gz to: {}'.format(job.commandfile))
            pickle.dump(job, fout)

        self._queue_job(job)

    def cancel_job(self, job):
        """
        Cancel a job previously queued

        :param job: job to cancel
        """
        if not isinstance(job, Job):
            try:
                job = self.job_dict[job]
            except KeyError:
                fastr.log.warning('Job {} is no longer under processing, cannot cancel!'.format(job))
                return

        fastr.log.debug('Cancelling {}'.format(job.jobid))

        job.status = self.job_status[job.id] = JobState.cancelled

        fastr.log.debug('Cancelling children for {}'.format(job.id))
        if job.id in self.held_queue:
            fastr.log.debug('Found children....')
            held_queue = self.held_queue[job.id]
            for dependent_job in held_queue:
                fastr.log.debug('Checking sub {}'.format(dependent_job))
                if dependent_job in self.job_dict and dependent_job in self.job_status and not self.job_status[dependent_job].done:
                    fastr.log.debug('Cancelling sub {}'.format(dependent_job))
                    self.cancel_job(dependent_job)
        else:
            fastr.log.debug('No children....')

        self._cancel_job(job)

        fastr.log.debug('Removing {} from jobdict'.format(job.jobid))
        self.job_archive[job.id] = job
        try:
            del self.job_dict[job.id]
        except KeyError:
            pass

        fastr.log.debug('Calling cancelled for {}'.format(job.id))
        if self._cancelled_callback is not None:
            self._cancelled_callback(job)

    def release_job(self, job):
        """
        Release a job that has been put on hold

        :param jobid: job to release
        """
        if not isinstance(job, Job):
            if job not in self.job_dict:
                fastr.log.warning('Job {} is no longer under processing, cannot release!'.format(job))
                return

            try:
                job = self.job_dict[job]
            except KeyError:
                fastr.log.warning('Job {} is no longer under processing, cannot release!'.format(job))
                return

        job.status = JobState.queued
        self._release_job(job.id)

    def job_finished(self, job, blocking=False):
        """
        The default callback that is called when a Job finishes. This will
        create a new thread that handles the actual callback.

        :param Job job: the job that finished
        :return:
        """
        if not blocking:
            # The callback has to finish immediately, so create thead to handle callback and return
            callback_thread = threading.Thread(target=self._job_finished_body,
                                               name='fastr_jobfinished_callback',
                                               args=(job,))
            callback_thread.start()
        else:
            self._job_finished_body(job)

    def _job_finished_body(self, job):
        """
        The actual callback that is executed in a separate thread. This
        method handles the collection of the result, the release of depending
        jobs and calling the user defined callback.

        :param Job job: the job that finished
        """
        fastr.log.debug('ExecutorInterface._job_finished_callback called')
        self.job_status[job.jobid] = JobState.processing_callback

        # The Job finished should always log the errors rather than
        # crashing the whole execution system
        # pylint: disable=bare-except
        try:
            try:
                job = iohelpers.load_gpickle(job.logfile)
            except EOFError:
                job.info_store['errors'].append(
                exceptions.FastrResultFileNotFound(
                    ('Could not read job result file {}, assuming '
                     'the job crashed during output write.').format(job.logfile)).excerpt())
                job.status = JobState.failed
            except IOError:
                job.info_store['errors'].append(
                    exceptions.FastrResultFileNotFound(
                        ('Could not find/read job result file {}, assuming '
                         'the job crashed before it created output.').format(job.logfile)).excerpt())
                job.status = JobState.failed

            if self._status_callback is not None:
                self._status_callback(job)
                job.status_callback = self._status_callback

        except:
            exc_type, _, trace = sys.exc_info()
            exc_info = traceback.format_exc()
            trace = traceback.extract_tb(trace, 1)[0]
            fastr.log.error('Encountered exception ({}) during execution:\n{}'.format(exc_type.__name__, exc_info))
            if 'errors' not in job.info_store:
                job.info_store['errors'] = []
            job.info_store['errors'].append((exc_type.__name__, exc_info, trace[0], trace[1]))
            job.status = JobState.execution_failed

        result = job
        fastr.log.debug('Finished {} with status {}'.format(job.jobid, job.status))
        jobid = result.jobid

        # Make sure the status is either finished or failed
        if result.status == JobState.execution_done:
            result.status = JobState.finished
        else:
            result.status = JobState.failed

        # Set the job status so the hold jobs will be release properly
        self.job_status[jobid] = result.status
        if jobid in self.job_dict:
            self.job_dict[jobid].status = JobState.finished

        # The ProcessPoolExecutor has to track job dependencies itself, so
        # therefor we have to check for jobs depending on the finished job
        if jobid in self.held_queue:
            fastr.log.debug('Signaling depending jobs {}'.format(self.held_queue[jobid]))
            ready_jobs = []
            for held_jobid in self.held_queue[jobid]:
                action = self.check_job_requirements(held_jobid)
                if action == JobAction.queue:
                    # Re-assign managed dict member
                    if held_jobid in self.job_dict:
                        held_job = self.job_dict[held_jobid]
                        self.job_dict[held_jobid] = held_job
                    fastr.log.debug('Job {} is now ready to be submitted'.format(held_jobid))

                    # If ready, flag job for removal from held queue and send
                    # to pool queue to be executed
                    ready_jobs.append(held_jobid)
                    self.release_job(held_jobid)
                elif action == JobAction.cancel:
                    fastr.log.debug('Job {} will be cancelled'.format(held_jobid))
                    ready_jobs.append(held_jobid)
                    self.cancel_job(held_jobid)
                else:
                    fastr.log.debug('Job {} still has unmet dependencies'.format(held_jobid))

            # Remove jobs that no longer need to be held from held_queue
            for readyjobid in ready_jobs:
                job = self.get_job(readyjobid)

                for hold_id in job.hold_jobs:
                    with self.held_queue_lock:
                        if hold_id in self.held_queue:
                            # remove from held_queue. because of the managed dict,
                            # we need to replace the value (not update)
                            required_job = self.held_queue[hold_id]
                            if readyjobid in required_job:
                                required_job.remove(readyjobid)
                            else:
                                fastr.log.warning('Could not remove {} from {}'.format(readyjobid, required_job))

            with self.held_queue_lock:
                del self.held_queue[jobid]

        # Extra subclass callback
        fastr.log.debug('Subclass callback')
        try:
            self._job_finished(result)
        except:
            exc_type, _, _ = sys.exc_info()
            exc_info = traceback.format_exc()
            fastr.log.error('Encountered exception ({}) during callback {}._job_finished:\n{}'.format(exc_type.__name__, type(self).__name__, exc_info))

        # Extra callback from object
        fastr.log.debug('Calling callback for {}'.format(jobid))
        if self._finished_callback is not None:
            try:
                self._finished_callback(result)
            except:
                if isinstance(self._finished_callback, functools.partial):
                    args = self._finished_callback.args + tuple('{}={}'.format(k, v) for k, v in self._finished_callback.keywords.items())
                    callback_name = '{f.__module__}.{f.func_name}({a})'.format(f=self._finished_callback.func, a=','.join(args))
                elif isinstance(self._finished_callback, types.FunctionType):
                    callback_name = '{f.__module__}.{f.func_name}'.format(f=self._finished_callback)
                elif isinstance(self._finished_callback, types.MethodType):
                    callback_name = '{m.__module__}.{m.im_class.__name__}.{m.im_func.func_name}'.format(m=self._finished_callback)
                else:
                    callback_name = repr(self._finished_callback)
                exc_type, _, _ = sys.exc_info()
                exc_info = traceback.format_exc()
                fastr.log.error('Encountered exception ({}) during callback {}:\n{}'.format(exc_type.__name__, callback_name, exc_info))
        else:
            fastr.log.debug('No callback specified')

        # Move the job to archive (to keep the number of working jobs limited
        # in the future the archive can be moved to a db/disk if needed
        fastr.log.debug('Archiving job {} with status {}'.format(jobid, result.status))
        try:
            del self.job_status[jobid]
        except KeyError:
            pass

        self.job_archive[jobid] = result

        try:
            del self.job_dict[jobid]
        except KeyError:
            pass
        fastr.log.debug('Done archiving')

    def get_job(self, jobid):
        try:
            return self.job_dict[jobid]
        except KeyError:
            try:
                return self.job_archive[jobid]
            except:
                raise exceptions.FastrKeyError('Could not find job {}'.format(jobid))

    def get_status(self, job):
        if not isinstance(job, Job):
            job = self.get_job(job)

        return job.status

    @abstractmethod
    def _queue_job(self, job):
        """
        Method that a subclass implements to actually queue a Job for execution

        :param job: job to queue
        """
        pass

    @abstractmethod
    def _cancel_job(self, jobid):
        """
        Method that a subclass implements to actually cancel a Job

        :param jobid: job to queue
        """
        pass

    @abstractmethod
    def _release_job(self, jobid):
        """
        Method that a subclass implements to actually release a job

        :param jobid: job to queue
        """
        pass

    @abstractmethod
    def _job_finished(self, job):
        """
        Method that a subclass can implement to add to the default callback.
        It will be called by ``_job_finished_body`` right before the user
        defined callback will be called.

        :param job: Job that resulted from the execution
        """
        pass

    def show_jobs(self, req_status=None):
        """
        List the queued jobs, possible filtered by status

        :param req_status: requested status to filter on
        :return: list of jobs
        """
        if isinstance(req_status, basestring):
            req_status = JobState[req_status]

        if not isinstance(req_status, JobState):
            return []

        results = []
        for key, status in self.job_status.items():
            if req_status is None or status == req_status:
                results.append(self.get_job(key))

        return results

    def check_job_status(self, jobid):
        """
        Get the status of a specified job

        :param jobid: the target job
        :return: the status of the job (or None if job not found)
        """
        try:
            return self.get_status(jobid)
        except exceptions.FastrKeyError:
            return None

    def check_job_requirements(self, jobid):
        """
        Check if the requirements for a job are fulfilled.

        :param jobid: job to check
        :return: directive what should happen with the job
        :rtype: JobAction
        """
        job = self.get_job(jobid)
        if job.hold_jobs is None or len(job.hold_jobs) == 0:
            return JobAction.queue

        all_done = True
        for hold_id in job.hold_jobs:
            status = self.check_job_status(hold_id)
            if status is not None and status != JobState.finished:
                if status.done:
                    return JobAction.cancel
                else:
                    fastr.log.debug('Dependency {} for {} is unmet ({})'.format(hold_id, jobid, status))
                    all_done = False

        if all_done:
            return JobAction.queue
        else:
            return JobAction.hold


class ExecutionPluginManager(PluginSubManager):
    """
    Container holding all the ExecutionPlugins known to the Fastr system
    """

    def __init__(self):
        """
        Initialize a ExecutionPluginManager and load plugins.

        :param path: path to search for plugins
        :param recursive: flag for searching recursively
        :return: newly created ExecutionPluginManager
        """
        super(ExecutionPluginManager, self).__init__(parent=fastr.plugin_manager,
                                                     plugin_class=self.plugin_class)

    @property
    def plugin_class(self):
        """
        The class of the Plugins expected in this BasePluginManager
        """
        return ExecutionPlugin



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABCMeta, abstractproperty

import sympy

from fastr import exceptions


class Dimension(object):
    """
    A class representing a dimension. It contains the name and size of the
    dimension.
    """
    def __init__(self, name, size):
        """
        The constructor for the dimension.

        :param str name: Name of the dimension
        :param size: Size fo the dimension
        :type size: int or sympy.Symbol
        """
        if not isinstance(name, str):
            raise exceptions.FastrTypeError("Dimension.name should be a str, "
                                            "found [{}] {}".format(type(name).__name__,
                                                                   size))

        if not isinstance(size, (int, sympy.Symbol)):
            raise exceptions.FastrTypeError("Dimension.size should be an int or"
                                            " sympy.Symbol, found [{}] {}".format(type(size).__name__,
                                                                                  size))

        if isinstance(size, int) and size < 0:
            raise exceptions.FastrValueError("Dimension.size should be non-negative")

        self.name = name
        self.size = size


class HasDimensions(object):
    """
    A Mixin class for any object that has a notion of dimensions and size. It
    uses the dimension property to expose the dimension name and size.
    """
    __metaclass__ = ABCMeta

    @abstractproperty
    def dimensions(self):
        """
        The dimensions has to be implemented by any subclass. It has to provide
        a tuple of Dimensions.

        :return: dimensions
        :rtype: tuple
        """

    @property
    def dimnames(self):
        """
        A tuple containing the dimension names of this object. All items of the
        tuple are of type str.
        """
        return tuple(x.name for x in self.dimensions)

    @property
    def size(self):
        """
        A tuple containing the size of this object. All items of the
        tuple are of type int or sympy.Symbol.
        """
        return tuple(x.size for x in self.dimensions)


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
import os
from collections import OrderedDict

import fastr
from fastr import exceptions
from fastr.core.baseplugin import Plugin

from fastr.core.pluginmanager import PluginSubManager
from fastr.core.interface import Interface, InterfaceResult, InputSpec, OutputSpec


class FlowPlugin(Plugin):
    """
    Plugin that can manage an advanced data flow. The plugins override the
    execution of node. The execution receives all data of a node in one go,
    so not split per sample combination, but all data on all inputs in one
    large payload. The flow plugin can then re-order the data and create
    resulting samples as it sees fits. This can be used for all kinds of
    specialized data flows, e.g. cross validation.

    To create a new FlowPlugin there is only one method that needs to be
    implemented: :py:meth:`execute <fastr.plugins.FlowInterface.flow_plugin_type.execute>`.
    """

    @abstractmethod
    def execute(payload):
        result_data = None
        log_data = None
        return result_data, log_data


class FlowPluginManager(PluginSubManager):
    """
    Container holding all the CollectorPlugins
    """

    def __init__(self):
        """
        Create the Coll
        :param path:
        :param recursive:
        :return:
        """
        super(FlowPluginManager, self).__init__(parent=fastr.plugin_manager,
                                                plugin_class=self.plugin_class)

    @property
    def plugin_class(self):
        """
        The class of the Plugins in the collection
        """
        return FlowPlugin

    @property
    def _instantiate(self):
        """
        Indicate that the plugins should instantiated before stored
        """
        return False


class FlowInterface(Interface):
    """
    The Interface use for AdvancedFlowNodes to create the advanced data flows
    that are not implemented in the fastr. This allows nodes to implement
    new data flows using the plugin system.

    The definition of ``FlowInterfaces`` are very similar to the default
    ``FastrInterfaces``.

    .. note::

        A flow interface should be using a specific FlowPlugin
    """

    __dataschemafile__ = 'FastrInterface.schema.json'

    flow_plugins = FlowPluginManager()
    flow_plugin_type = FlowPlugin

    def __init__(self, id_, document):
        super(FlowInterface, self).__init__()

        # Load from file if it is not a dict
        if not isinstance(document, dict):
            fastr.log.debug('Trying to load file: {}'.format(document))
            filename = os.path.expanduser(document)
            filename = os.path.abspath(filename)
            document = self._loadf(filename)
        else:
            document = self.get_serializer().instantiate(document)

        #: The ID of the interface
        self.id = id_

        # Create the inputs/outputs spec to expose to the rest of the system
        self._inputs = OrderedDict((v['id'], InputSpec(id_=v['id'],
                                                       cardinality=v.get('cardinality', 1),
                                                       datatype=v['datatype'] if 'datatype' in v else fastr.typelist.create_enumtype('__{}__{}__Enum__'.format(self.id, v['id']),
                                                                                                                                     tuple(v['enum'])).id,
                                                       required=v.get('required', True),
                                                       description=v.get('description', ''),
                                                       default=v.get('default', None),
                                                       hidden=v.get('hidden', False))) for v in document['inputs'])

        self._outputs = OrderedDict((v['id'], OutputSpec(id_=v['id'],
                                                         cardinality=v.get('cardinality', 1),
                                                         datatype=v['datatype'] if 'datatype' in v else fastr.typelist.create_enumtype('__{}__{}__Enum__'.format(self.id, v['id']),
                                                                                                                                       tuple(v['enum'])).id,
                                                         automatic=v.get('automatic', False),
                                                         required=v.get('required', True),
                                                         description=v.get('description', ''),
                                                         hidden=v.get('hidden', False))) for v in document['outputs'])

    def __eq__(self, other):
        if not isinstance(other, FlowInterface):
            return NotImplemented

        return vars(self) == vars(other)

    def __getstate__(self):
        """
        Get the state of the FastrInterface object.

        :return: state of interface
        :rtype: dict
        """
        state = {
            'id': self.id,
            'class': type(self).__name__,
            'inputs': self.inputs.values(),
            'outputs': self.outputs.values(),
        }

        return state

    def __setstate__(self, state):
        """
        Set the state of the Interface
        """
        self.id = state['id']

        self._inputs = OrderedDict((v['id'], InputSpec(id_=v['id'],
                                                       cardinality=v.get('cardinality', 1),
                                                       datatype=v['datatype'] if 'datatype' in v else fastr.typelist.create_enumtype('__{}__{}__Enum__'.format(self.id, v['id']),
                                                                                                                                     tuple(v['enum'])).id,
                                                       required=v.get('required', True),
                                                       description=v.get('description', ''),
                                                       default=v.get('default', None),
                                                       hidden=v.get('hidden', False))) for v in state['inputs'])

        self._outputs = OrderedDict((v['id'], OutputSpec(id_=v['id'],
                                                         cardinality=v.get('cardinality', 1),
                                                         datatype=v['datatype'] if 'datatype' in v else fastr.typelist.create_enumtype('__{}__{}__Enum__'.format(self.id, v['id']),
                                                                                                                                       tuple(v['enum'])).id,
                                                         automatic=v.get('automatic', False),
                                                         required=v.get('required', True),
                                                         description=v.get('description', ''),
                                                         hidden=v.get('hidden', False))) for v in state['outputs'])

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    def expanding(self):
        return 1

    def execute(self, target, payload):
        try:
            flow_plugin = self.flow_plugins[target.binary]
        except KeyError:
            raise exceptions.FastrKeyError('Could not find {} in {} (options {})'.format(target.binary, self.flow_plugins, self.flow_plugins.keys()))
        result = InterfaceResult(result_data=None, log_data=None, payload=payload)
        result.result_data, result.log_data = flow_plugin.execute(payload)

        return result



#!/usr/bin/env python


def create_network():
    # Import the faster environment and set it up
    import fastr
    # Create a new network
    network = fastr.Network(id_='collapse')

    # Create a source node in the network
    source = network.create_source(fastr.typelist['Int'], id_='source')

    # Create a sink to save the data
    sink = network.create_sink(fastr.typelist['Int'], id_='sink')

    # Link the addint node to the sink
    link = network.create_link(source.output, sink.input)
    link.collapse = 0

    return network


def source_data(network):
    return {
        'source': {
            'sample_a': (1, 2, 3, 4),
            'sample_b': (5, 6, 7, 8),
            'sample_c': (9, 10, 11, 12),
        }
    }


def sink_data(network):
    return {'sink': 'vfs://tmp/results/{}/result_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id)}


def main():
    network = create_network()

    # Execute
    network.draw_network()
    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module containing the class that represent versions
"""

import re
import fastr.exceptions as exceptions


class Version(tuple):
    """
    Class representing a software version definition. Allows for sorting and
    extraction of parts.
    """
    # Regular expression to split a version into 7 fields: major, minor, extras, seperator, status, build, and suffix
    version_matcher = re.compile(r'(\d+)\.(\d+)((?:\.\d+)+)?([_\-\.])?(a(?=\d)|b(?=\d)|alpha(?=\d)|beta(?=\d)|rc(?=\d)|r(?=\d))?(\d+)?([a-zA-Z0-9\-_\.]*)')

    def __new__(cls, *version):
        """Class containing a version

        Can be constructed by:

        .. code-block:: python

          Version( 'major.$minor.$extra[0].$extra[1]$seperator$status$build$suffix' )
          Version( major, minor, extra, status, build, suffix, seperator )
          Version( (major, minor, extra, status, build, suffix, seperator) )
          Version( [major, minor, extra, status, build, suffix, seperator] )

        :param int major: interger giving major version
        :param int minor: is an integer (required)
        :param extra: is a list of integers
        :type extra: list of int
        :param str status: can be "a", "alpha", "b", "beta", "rc", or "r"
        :param int build: is an integer
        :param str suffix: can contain any combination of alpha-numeric character and "._-"
        :param str seperator: is any of ".", "-", or "_", which is located between $extra and $build

        .. note::

           The method based on strings is the recommended method. For strings
           the major and minor version are required, where for tuple and list
           constructors all seven elements are optional.

        Examples:

        .. code-block:: python

           >>> a = Version('0.1')
           >>> print(tuple(a))
           (0, 1, None, None, None, '', None)
           >>> b = Version('2.5.3-rc2')
           >>> print(tuple(b))
           (2, 5, [3], 'rc', 2, '', '-')
           >>> c = Version('1.2.3.4.5.6.7-beta8_with_suffix')
           >>> print(tuple(c))
           (1, 2, [3, 4, 5, 6, 7], 'beta', 8, '_with_suffix', '-')


        """
        # Check if arguments are a list of arguments or a single argument containing a list, tuple or str
        if len(version) == 1:
            if (isinstance(version[0], list) or
                    isinstance(version[0], tuple) or
                    isinstance(version[0], str) or
                    isinstance(version[0], unicode)):
                version = version[0]
            else:
                raise exceptions.FastrVersionInvalidError('"{}" is not a valid version input!'.format(version))

        orig_version = repr(version)
        if isinstance(version, unicode):
            version = str(version)

        if isinstance(version, str):
            version = cls.version_matcher.match(version)
            if version is None:
                raise exceptions.FastrVersionInvalidError('"{}" is not a valid version input!'.format(orig_version))

            version = version.groups()

            # Move seperator (element at index 3) to the end
            version = list(version)
            version.append(version[3])
            version.pop(3)

        if isinstance(version, tuple):
            version = list(version)

        if not isinstance(version, list) or len(version) > 7:
            raise exceptions.FastrVersionInvalidError('"{}" is not a valid version input!'.format(orig_version))

        default = [0, 0, None, None, None, None, None]

        if len(version) < 7:
            version += default[len(version):]

        # Parse the fields
        version[0] = int(version[0])
        version[1] = int(version[1])
        if isinstance(version[2], str):
            version[2] = re.split(r'\.', version[2].strip('.'))
            version[2] = tuple(int(x) for x in version[2])
        if isinstance(version[2], list):
            version[2] = tuple(version[2])
        # Small hack to make sure r > rc
        if version[3] == 'r':
            version[3] = 're'
        if version[4] is not None:
            version[4] = int(version[4])

        return super(Version, cls).__new__(Version, version)

    def __str__(self):
        """
        Return a string representation of the version
        """
        string = '{v.major}.{v.minor}{v.extra_string}'.format(v=self)

        if self[6] is not None:
            string += str(self[6])

        if self.status is not None:
            string += str(self.status)

        if self.build is not None:
            string += str(self.build)

        if self.suffix is not None:
            string += str(self.suffix)

        return string

    def __repr__(self):
        """
        Return a in-editor representation of the version
        """
        return '<Version: ' + str(self) + '>'

    @property
    def major(self):
        """
        major version
        """
        return self[0]

    @property
    def minor(self):
        """
        minor version
        """
        return self[1]

    @property
    def extra(self):
        """
        extra version extension as a list
        """
        return self[2]

    @property
    def extra_string(self):
        """
        extra version extension as a string
        """
        if self[2] is None:
            return ''
        else:
            return ''.join(['.{}'.format(x) for x in self[2]])

    @property
    def status(self):
        """
        the status of the version (a, alpha, b, beta, rc or r)
        """
        if self[3] == 're':
            return 'r'
        else:
            return self[3]

    @property
    def build(self):
        """
        the build number, this is following the status (e.g. for
        3.2-beta4, this would be 4)
        """
        return self[4]

    @property
    def suffix(self):
        """
        the remainder of the version which was not formatted in a known way
        """
        return self[5]



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
FASTR_LOG_TYPE = 'none'  # Do not get info about fastr


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('resource_type', metavar='TYPE', choices=['tool'],
                        help='Type of resource to verify (e.g. tool)')
    parser.add_argument('path', help='path of the resource to verify')
    return parser


def main(args=None, unknown_args=None):
    """
    Print information from a job file
    """
    if args is None and unknown_args is None:
        # No arguments were parsed yet, parse them now
        parser = get_parser()
        args, unknown_args = parser.parse_known_args()

    import fastr
    from fastr.utils.verify import verify_tool

    resource_type = args.resource_type.lower()
    if resource_type == 'tool':
        verify_tool(args.path)
    else:
        fastr.log.error('Unknown resource type {}'.format(resource_type))


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module providing tools to parse and create valid urls and paths.

usage example:

When in fastr.config under the mounts section the data mount is set to /media/data, you will get the following.
.. code-block:: python

  >>> import fastrdata.url import get_path_from_url
  >>> get_path_from_url('vfs://data/temp/blaat1.png')
  '/media/data/temp/blaat1.png'
"""

import re
import os.path
import urlparse as up
import posixpath
import fastr
from fastr.exceptions import FastrUnknownURLSchemeError


def register_url_scheme(scheme):
    """ Register a custom scheme to behave http like. This is needed to parse
    all things properly.
    """
    for method in [s for s in dir(up) if s.startswith('uses_')]:
        getattr(up, method).append(scheme)


def get_url_scheme(url):
    """
    Get the schem of the url

    :param str url: url to extract scheme from
    :return: the url scheme
    :rtype: str
    """
    parsed_url = up.urlparse(str(url))
    return parsed_url.scheme


def get_path_from_url(url):
    """ Get the path to a file from a url.
    Currently supports the file:// and vfs:// scheme's

    Examples:

    .. code-block:: python

      >>> url.get_path_from_url('vfs://neurodata/user/project/file.ext')
      'Y:\\neuro3\\user\\project\\file.ext'


      >>> 'file:///d:/data/project/file.ext'
      'd:\\data\\project\\file.ext'

    .. warning::

      file:// will not function cross platform and is mainly for testing

    """
    # Translate properly depending on the scheme being used
    scheme = get_url_scheme(url)

    # Make the paths and vfs not go via ioplugins, but shortcut it
    if scheme == '':
        # This is not a URL, but a path, throw error
        raise FastrUnknownURLSchemeError('URL using no url scheme in {}, so it must be a path! '.format(url))
    elif scheme == 'vfs':
        # Directly call vfs
        return fastr.vfs.url_to_path(url)
    else:
        # Dispatch to ioplugin via the ioplugin manager
        try:
            ioplugin = fastr.ioplugins[scheme]
        except KeyError:
            raise FastrUnknownURLSchemeError('URL using an unknown scheme in {}'
                                             ' ({} not in {})'.format(url,
                                                                      scheme,
                                                                      fastr.ioplugins.keys()))

        return ioplugin.url_to_path(url)


def isurl(string):
    """
    Check if string is a valid url

    :param str string: potential url
    :return: flag indicating if string is a valid url
    """
    if not isinstance(string, (str, unicode)):
        return False

    parsed_url = up.urlparse(str(string))
    return len(parsed_url.scheme) > 1


def basename(url):
    """
    Get basename of url

    :param str url: the url
    :return: the basename of the path in the url
    """
    parsed_url = up.urlparse(str(url))
    return posixpath.basename(parsed_url.path)


def dirname(url):
    """
    Get the dirname of the url

    :param str url: the url
    :return: the dirname of the path in the url
    """
    parsed_url = up.urlparse(str(url))
    return posixpath.dirname(parsed_url.path)


def dirurl(url):
    """
    Get the a new url only having the dirname as the path

    :param str url: the url
    :return: the modified url with only dirname as path
    """
    return split(url)[0]


def split(url):
    """
    Split a url in a url with the dirname and the basename part of the path of
    the url

    :param str url: the url
    :return: a tuple with (dirname_url, basename)
    """
    # pylint: disable=protected-access
    parsed_url = up.urlparse(str(url))._asdict()
    parsed_url['path'], part = posixpath.split(parsed_url['path'])
    return up.urlunparse(parsed_url.values()), part


def join(url, *p):
    """
    Join the path in the url with p

    :param str url: the base url to join with
    :param p: additional parts of the path
    :return: the url with the parts added to the path
    """
    # pylint: disable=protected-access
    parsed_url = up.urlparse(str(url))._asdict()
    parsed_url['path'] = posixpath.join(parsed_url['path'], *p)
    return up.urlunparse(parsed_url.values())


def normurl(url):
    """
    Normalized the path of the url

    :param str url: the url
    :return: the normalized url
    """
    # pylint: disable=protected-access
    parsed_url = up.urlparse(str(url))._asdict()
    parsed_url['path'] = posixpath.normpath(parsed_url['path'])
    return up.urlunparse(parsed_url.values())


def create_vfs_url(mountpoint, path):
    """
    Construct an url from a given mount point and a relative path to the mount point.

    :param str mountpoint: the name of the mountpoint
    :param str path: relative path from the mountpoint
    :return: the created vfs url
    """
    return "vfs://%s/%s" % (mountpoint, path.strip('/'))


def full_split(urlpath):
    """
    Split the path in the url in a list of parts

    :param urlpath: the url path
    :return: a list of parts
    """
    parts = []
    while True:
        newpath, tail = posixpath.split(urlpath)
        if newpath == urlpath:
            assert not tail
            if urlpath:
                parts.append(urlpath)
            break
        parts.append(tail)
        urlpath = newpath
    parts.reverse()
    return parts


def _correct_separators(path):
    """
    Translates the URL seperator '/' into the apropriate seperator for the OS

    :param str path: the path to correct
    :return: path with consistent separators
    """
    return re.sub('/', ('%r' % os.path.sep).strip("'"), path.strip('/'))



__author__ = 'hachterberg'

import re

import fastr
from fastr.plugins import FastrInterface


class StdoutCollector(FastrInterface.collector_plugin_type):
    """
    The StdoutCollector can collect data from the stdout stream of a program.
    It filters the ``stdout`` line by line matching a predefined regular expression.

    The general working is as follows:

    1. The location field is taken from the output
    2. The substitutions are performed on the location field (see below)
    3. The updated location field will be used as a :ref:`regular expression <python:re-syntax>` filter
    4. The ``stdout`` is scanned line by line and the :ref:`regular expression <python:re-syntax>` filter is applied

    The special substitutions performed on the location use the
    Format Specification Mini-Language :ref:`python:formatspec`.
    The predefined fields that can be used are:

    * ``inputs``, an objet with the input values (use like ``{inputs.image[0]}``)
    * ``outputs``, an object with the output values (use like ``{outputs.result[0]}``)
    * ``special`` which has two subfields:

      * ``special.cardinality``, the index of the current cardinality
      * ``special.extension``, is the extension for the output DataType

    .. note:: because the plugin scans line by line, it is impossible to catch
              multi-line output into a single value
    """
    def __init__(self):
        super(StdoutCollector, self).__init__()
        self.id = 'stdout'

    def _collect_results(self, interface, output, result):
        output_data = []
        result.result_data[output.id] = output_data
        location = output.location
        stdout = result.log_data['stdout']
        cardinality = 0
        fastr.log.debug('Searching for {}'.format(location))
        specials, inputs, outputs, input_parts, outputs_parts = interface.get_specials(result.payload, output, cardinality)
        loc = location.format(input=inputs, output=outputs, special=specials, input_parts=input_parts, output_parts=outputs_parts)
        for line in stdout.splitlines():
            match = re.search(loc, line)
            if match is not None:
                value = '{regexp[0]}'.format(input=inputs, output=outputs, special=specials, input_parts=input_parts, output_parts=outputs_parts, regexp=match.groups())

                if output.separator is not None:
                    value = value.split(output.separator)
                else:
                    value = [value]

                for val in value:
                    fastr.log.info('Found value: {}'.format(val))
                    output_data.append(val)
                    cardinality += 1

                specials, inputs, outputs, input_parts, outputs_parts = interface.get_specials(result.payload, output, cardinality)
                loc = location.format(input=inputs, output=outputs, special=specials, input_parts=input_parts, output_parts=outputs_parts)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastr
from fastr.datatypes import ValueType


class UnsignedInt(ValueType):
    description = 'an unsigned integer value'

    def _validate(self):
        """
        Validate the value of the DataType.

        :return: flag indicating validity of the Int
        :rtype: bool
        """
        if self.value is None:
            return False
        elif isinstance(self.value, int):
            return self.value >= 0
        else:
            return False

    @property
    def value(self):
        """
        The value of object instantiation of this DataType.
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Setter function for value property
        """
        try:
            self._value = int(value)
        except (ValueError, TypeError):
            self._value = None
            fastr.log.debug('Not a valid value for a Int ({}), ignoring!'.format(value))




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some helper function to compare dictionaries and find the parts of the dict
that are different. This is mostly to help in debugging.
"""
import fastr
from fastr.core.serializable import Serializable


def dicteq(self, other):
    """
    Compare two dicts for equality

    :param self: the first object to compare
    :param other: the oth
    :return:
    """
    if not fastr.config.debug:
        return self == other

    if self == other:
        return True

    fastr.log.debug('Differences:\n{}'.format('\n'.join(diffdict(self, other))))
    return False


def diffdict(self, other, path=None, visited=None):
    """
    Find the differences in two dictionaries.

    :param self: the first object to compare
    :param dict other: other dictionary
    :param list path: the path for nested dicts (too keep track of recursion)
    :return: list of messages indicating the differences
    :rtype: list
    """
    if path is None:
        path = []

    if visited is None:
        visited = []

    messages = []

    for k in self:
        if k not in other:
            messages.append('Key "{}" type "{}" (/{}) not found in other'.format(k,
                                                                                 type(k).__name__,
                                                                                 '/'.join(path)))

    for k in other:
        if k not in self:
            messages.append('Key "{}" type "{}" (/{}) not found in self'.format(k,
                                                                                type(k).__name__,
                                                                                '/'.join(path)))
        else:
            if self[k] != other[k]:
                if type(self[k]) is not type(other[k]):
                    messages.append('Values for "{}" (/{}) have different type ({} vs {})'.format(k,
                                                                                                  '/'.join(path),
                                                                                                  type(self[k]).__name__,
                                                                                                  type(other[k]).__name__))
                elif isinstance(self[k], dict):
                    messages.extend(diffdict(self[k], other[k], path + [k], visited=visited))
                elif isinstance(self[k], Serializable):
                    messages.extend(diffobj(self[k], other[k], path + [k], visited=visited))
                else:
                    messages.append('Values for "{}" (/{}) differ ({} vs {})'.format(k,
                                                                                     '/'.join(path),
                                                                                     self[k],
                                                                                     other[k]))

    return messages


def diffobj(self, other, path=None, visited=None):
    """
    Compare two objects by comparing their __dict__ entries

    :param self: the first object to compare
    :param other: other objects to compare
    :param list path: the path for nested dicts (too keep track of recursion)
    :return: list of messages
    :rtype: list
    """
    if type(self) is not type(other):
        return ['The types of the object do not match!']

    if visited is None:
        visited = []

    if (id(self), id(other)) not in visited:
        visited.append((id(self), id(other)))
        return diffdict(vars(self), vars(other), path=path, visited=visited)
    else:
        return []


def diffobj_str(self, other):
    """
    Compare two objects by comparing their __dict__ entries, but returns the
    differences in a single string ready for logging.

    :param self: the first object to compare
    :param other: other object to compare to
    :return: the description of the differences
    :rtype: str
    """
    return 'Differences:\n{}'.format('\n'.join(diffobj(self, other)))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains a number of functions for checksumming files and objects
"""
import hashlib
import os
import pickle
from fastr import exceptions


def checksum(filepath, algorithm='md5', hasher=None):
    """
    Generate the checksum of a file

    :param filepath: path of the file(s) to checksum
    :type filepath: str, list
    :param str algorithm: the algorithm to use
    :param _hashlib.HASH hasher: a hasher to continue updating (rather then creating a new one)
    :return: the checksum
    :rtype: str
    """
    if isinstance(filepath, str):
        filepath = [filepath]

    if hasher is None:
        hasher = hashlib.new(algorithm)

    for path in filepath:
        if os.path.isdir(path):
            checksum_directory(path, hasher=hasher)
        else:
            with open(path, 'rb') as file_handle:
                while True:
                    data = file_handle.read(32768)
                    if not data:
                        break
                    hasher.update(data)
    return hasher.hexdigest()


def checksum_directory(directory, algorithm='md5', hasher=None):
    """
    Generate the checksum of an entire directory

    :param str directory: path of the file(s) to checksum
    :param str algorithm: the algorithm to use
    :param _hashlib.HASH hasher: a hasher to continue updating (rather then creating a new one)
    :return: the checksum
    :rtype: str
    """
    if not os.path.isdir(directory):
        raise exceptions.FastrValueError('{} does not point to a valid directory!')

    if hasher is None:
        hasher = hashlib.new(algorithm)

    for root, dirs, files in os.walk(directory):
        relroot = os.path.relpath(root, directory)
        for d in dirs:
            hasher.update(os.path.join(relroot, d))

        for filename in files:
            hasher.update(os.path.join(relroot, filename))

            with open(os.path.join(root, filename)) as file_handle:
                while True:
                    data = file_handle.read(32768)
                    if not data:
                        break
                    hasher.update(data)

    return hasher.hexdigest()


def hashsum(objects, hasher=None):
    """
    Generate the md5 checksum of (a) python object(s)

    :param objects: the objects to hash
    :param hasher: the hasher to use as a base
    :return: the hash generated
    :rtype: str
    """
    if hasher is None:
        hasher = hashlib.new('md5')

    if not isinstance(objects, list):
        objects = [objects]

    for obj in objects:
        if isinstance(obj, (str, unicode)):
            hasher.update(obj)
        elif isinstance(obj, (bool, int, float)):
            hasher.update(str(obj))
        elif isinstance(obj, file):
            obj.seek(0)  # Make sure to hash entire file
            while True:
                data = obj.read(32768)
                if not data:
                    break
                hasher.update(data)
        elif isinstance(obj, list):
            hasher.update('_||LIST||_')
            for element in obj:
                hashsum(element, hasher=hasher)
                hasher.update('_|ENDELEM|_')
            hasher.update('_||ENDLIST||_')
        elif isinstance(obj, tuple):
            hasher.update('_||TUPLE||_')
            for element in obj:
                hashsum(element, hasher=hasher)
                hasher.update('_|ENDELEM|_')
            hasher.update('_||ENDTUPLE||_')
        elif isinstance(obj, dict):
            hasher.update('_||DICT||_')
            for key in sorted(obj.keys()):
                hashsum(key, hasher=hasher)
                hasher.update('_|ENDKEY|_')
                hashsum(obj[key], hasher=hasher)
                hasher.update('_|ENDVALUE|_')
            hasher.update('_||ENDDICT||_')
        elif isinstance(obj, set):
            hasher.update('_||SET||_')
            for element in sorted(obj):
                hashsum(element, hasher=hasher)
                hasher.update('_|ENDELEM|_')
            hasher.update('_||ENDSET||_')
        else:
            hasher.update(pickle.dumps(obj))

    return hasher.hexdigest()


def md5_checksum(filepath):
    """
    Generate the md5 checksum of a file

    :param filepath: path of the file(s) to checksum
    :type filepath: str, list
    :return: the checksum
    :rtype: str
    """
    return checksum(filepath)


def sha1_checksum(filepath):
    """
    Generate the sha1 checksum of a file

    :param filepath: path of the file(s) to checksum
    :type filepath: str, list
    :return: the checksum
    :rtype: str
    """
    return checksum(filepath, 'sha1')
